from abc import ABC, abstractmethod
from typing import Generic, TypeVar
from sqlglot import Dialects, expressions as exp

from ...schema import Reference, Table, Column, Index, Enum


DBMLNode = TypeVar("DBMLNode", Table, Column, Index, Enum, Reference)


class BaseSQLConverter(Generic[DBMLNode], ABC):
    def __init__(self, dialect: Dialects):
        self.dialect = dialect

    @abstractmethod
    def convert(self, node: DBMLNode) -> exp.Expression:
        raise NotImplementedError("conver function is not implemented")



from sqlglot import expressions as exp

from ...schema import Reference
from .base import BaseSQLConverter


class ReferenceConverter(BaseSQLConverter[Reference]):
    def convert(self, node):
        reference = node
        options = []
        if reference.settings:
            if reference.settings.delete:
                options.append(f"ON DELETE {reference.settings.delete.upper()}")
            if reference.settings.update:
                options.append(f"ON UPDATE {reference.settings.update.upper()}")
        return exp.Constraint(
            this=(
                exp.Identifier(this=reference.name, quoted=False)
                if reference.name
                else None
            ),
            expressions=[
                exp.ForeignKey(
                    expressions=list(
                        map(
                            lambda column: exp.Identifier(this=column, quoted=True),
                            reference.from_columns,
                        )
                    ),
                    reference=exp.Reference(
                        this=exp.Schema(
                            this=exp.Table(
                                this=exp.Identifier(
                                    this=reference.to_table.name, quoted=True
                                ),
                                db=exp.Identifier(
                                    this=reference.to_table.db_schema, quoted=True
                                )
                                if reference.to_table.db_schema
                                else None,
                            ),
                            expressions=list(
                                map(
                                    lambda column: exp.Identifier(
                                        this=column, quoted=True
                                    ),
                                    reference.to_columns,
                                )
                            ),
                        ),
                        options=options,
                    ),
                )
            ],
        )



from typing import Any, Annotated, Literal, List
from pydantic import BaseModel, ConfigDict, Field, BeforeValidator
from pydantic.aliases import AliasChoices


class Settings(BaseModel):
    model_config = ConfigDict(extra="allow")
    note: str | None = None


class Noteable(BaseModel):
    model_config = ConfigDict(extra="allow")
    note: str | None = Field(
        default=None, validation_alias=AliasChoices("note", "Note")
    )


class Name(BaseModel):
    db_schema: str | None = None
    name: str | None = None


# Project
class Project(Name, Noteable):
    database_type: str | None = None


# TableGroup
class TableGroup(Name, Noteable):
    tables: List[Name]
    settings: Settings | None = None


# Enum
class EnumValue(BaseModel):
    value: str
    settings: Settings | None = None


class Enum(Name):
    values: List[EnumValue]


# Sticky Note
class Note(Noteable):
    pass


# Relationship
class ReferenceInline(BaseModel):
    relationship: Literal["-", ">", "<", "<>"]
    to_table: Name
    to_columns: Annotated[
        List[str], BeforeValidator(lambda v: v if not v or isinstance(v, List) else [v])
    ] = None


class ReferenceSettings(Settings):
    delete: (
        Literal["cascade", "restrict", "set null", "set default", "no action"] | None
    ) = None
    update: (
        Literal["cascade", "restrict", "set null", "set default", "no action"] | None
    ) = None
    color: str | None = None  # For rendering


class Reference(Name, ReferenceInline):
    settings: ReferenceSettings | None = None
    from_table: Name | None = None
    from_columns: Annotated[
        List[str], BeforeValidator(lambda v: v if not v or isinstance(v, List) else [v])
    ] = None


# Table
class DataType(BaseModel):
    sql_type: str
    length: int | None = None
    scale: int | None = None


class ColumnSettings(Settings):
    is_primary_key: bool = False
    is_null: bool = True
    is_unique: bool = False
    is_increment: bool = False
    default: Any | None = None
    ref: ReferenceInline | None = None


class Column(Name):
    data_type: DataType | Name
    settings: ColumnSettings | None = None


class IndexSettings(Settings):
    idx_type: Literal["btree", "hash"] | None = Field(default=None, alias="type")
    name: str | None = None
    is_unique: bool = False
    is_primary_key: bool = False


class Index(BaseModel):
    columns: Annotated[
        List[str], BeforeValidator(lambda v: v if not v or isinstance(v, List) else [v])
    ] = None
    settings: IndexSettings | None = None


class TableSettings(Settings):
    header_color: str | None = Field(
        default=None, validation_alias=AliasChoices("headercolor", "headerColor")
    )


class TablePartial(Name):
    columns: List[Column]
    indexes: List[Index] | None = None
    settings: TableSettings | None = None


class Table(TablePartial, Noteable):
    alias: str | None = None
    table_partials: List[str] | None = None


# Diagram
class Diagram(BaseModel):
    project: Project | None = None
    enums: List[Enum] | None = []
    table_groups: List[TableGroup] | None = []
    sticky_notes: List[Note] | None = []
    references: List[Reference] | None = []
    tables: List[Table] | None = []
    table_partials: List[TablePartial] | None = []



from sqlglot import Dialects, expressions as exp

from ...schema import Index, Table
from .base import BaseSQLConverter


class IndexConverter(BaseSQLConverter[Index]):
    def __init__(self, dialect, table: Table):
        super().__init__(dialect)
        self.table = table

    @property
    def support_nulls_first(self) -> bool:
        return self.dialect in [
            Dialects.POSTGRES,
            Dialects.ORACLE,
            Dialects.DUCKDB,
        ]

    def convert(self, node):
        index = node
        columns = []
        for column in index.columns:
            is_func_exp = "`" in column
            name = f"({column.strip('`')})" if is_func_exp else column
            columns.append(
                exp.Ordered(
                    this=exp.Column(
                        this=exp.Identifier(this=name, quoted=not is_func_exp)
                    ),
                    nulls_first=not self.support_nulls_first,
                )
            )

        index_ref = exp.Create(
            this=exp.Index(
                this=exp.Identifier(this=index.settings.name, quoted=True)
                if index.settings and index.settings.name
                else None,
                table=exp.Table(
                    this=exp.Identifier(this=self.table.name, quoted=True),
                    db=exp.Identifier(this=self.table.db_schema, quoted=True)
                    if self.table.db_schema
                    else None,
                ),
                params=exp.IndexParameters(
                    columns=columns,
                    using=exp.Var(this=index.settings.idx_type.upper())
                    if index.settings and index.settings.idx_type
                    else None,
                ),
            ),
            kind="INDEX",
            unique=index.settings and index.settings.is_unique,
        )
        return index_ref


class CompositePKIndexConverter(BaseSQLConverter[Index]):
    def convert(self, node):
        index = node
        return exp.PrimaryKey(
            expressions=list(
                map(
                    lambda column: exp.Identifier(this=column, quoted=True),
                    index.columns,
                )
            )
        )



from lark_dbml import load
from lark_dbml.converter import to_sql


def test_sql(example_path, expectation_path):
    diagram = load(example_path / "complex.dbml")

    sql = to_sql(diagram)

    with open(expectation_path / "complex_postgres.sql") as f:
        expectation = f.read()
    assert sql == expectation



from typing import List
from sqlglot import Dialects, expressions as exp
from .base import BaseSQLConverter
from ...schema import Column, ColumnSettings, DataType, Enum


class ColumnConverter(BaseSQLConverter[Column]):
    def __init__(self, dialect, enums: List[Enum]):
        super().__init__(dialect)
        self.enums = enums

    @property
    def support_inline_enum_def(self) -> bool:
        return self.dialect == Dialects.MYSQL

    def _convert_settings(self, settings: ColumnSettings) -> List[exp.Constraint]:
        constraints = []
        if settings.is_increment:
            constraints.append(
                exp.ColumnConstraint(
                    kind=exp.GeneratedAsIdentityColumnConstraint(this=True)
                )
            )
        if settings.is_unique:
            constraints.append(exp.ColumnConstraint(kind=exp.UniqueColumnConstraint()))
        if settings.is_primary_key:
            constraints.append(
                exp.ColumnConstraint(kind=exp.PrimaryKeyColumnConstraint())
            )
        # Only set NULL or NOT NULL if the column is not a primary key
        if not (settings.is_primary_key or settings.is_increment or settings.is_unique):
            constraints.append(
                exp.ColumnConstraint(
                    kind=exp.NotNullColumnConstraint(allow_null=settings.is_null)
                )
            )
        # Default value
        if settings.default:
            is_func_exp = isinstance(settings.default, str) and "`" in settings.default
            default_value = (
                settings.default if not is_func_exp else settings.default.strip("`")
            )
            constraints.append(
                exp.ColumnConstraint(
                    kind=exp.DefaultColumnConstraint(
                        this=exp.Literal(
                            this=default_value,
                            is_string=isinstance(default_value, str)
                            and not is_func_exp,
                        )
                    )
                )
            )
        # Reference
        if settings.ref:
            constraints.append(
                exp.ColumnConstraint(
                    kind=exp.Reference(
                        this=exp.Schema(
                            this=exp.Table(
                                this=exp.Identifier(
                                    this=settings.ref.to_table.name, quoted=True
                                ),
                                db=exp.Identifier(
                                    this=settings.ref.to_table.db_schema, quoted=True
                                )
                                if settings.ref.to_table.db_schema
                                else None,
                            ),
                            expressions=[
                                exp.Identifier(
                                    this=settings.ref.to_columns[0], quoted=True
                                )
                            ],
                        )
                    )
                )
            )
        return constraints

    def convert(self, node):
        column = node

        data_type: exp.DataType.Type = exp.DataType.Type.UNKNOWN
        enum: Enum = None
        if isinstance(column.data_type, DataType):
            type_name = column.data_type.sql_type
        else:
            type_name = f"{column.data_type.db_schema}.{column.data_type.name}"

        try:
            data_type = exp.DataType.Type(type_name.upper())
        except Exception:
            for e in self.enums:
                enum_name = e.name if not e.db_schema else f"{e.db_schema}.{e.name}"
                if type_name == enum_name:
                    enum = e
                    data_type = exp.DataType.Type.ENUM
                    break
            else:
                data_type = exp.DataType.Type.USERDEFINED

        if data_type == exp.DataType.Type.ENUM and not self.support_inline_enum_def:
            data_type = exp.DataType.Type.USERDEFINED

        match data_type:
            case exp.DataType.Type.ENUM:
                kind = exp.DataType(
                    this=data_type,
                    nested=False,
                    expressions=list(
                        map(
                            lambda value: exp.Literal(this=value.value, is_string=True),
                            enum.values,
                        )
                    ),
                )
            case exp.DataType.Type.USERDEFINED:
                kind = exp.DataType(
                    this=data_type,
                    nested=False,
                    kind=(type_name if " " not in type_name else f'"{type_name}"'),
                )
            case _:
                kind = exp.DataType(this=data_type, nested=False)

        col_def = exp.ColumnDef(
            this=exp.Identifier(this=column.name, quoted=True), kind=kind
        )

        if column.settings:
            col_def.set("constraints", self._convert_settings(column.settings))
        return col_def



from sqlglot import expressions as exp

from ...schema import Enum
from .base import BaseSQLConverter


class EnumConverter(BaseSQLConverter[Enum]):
    def convert(self, node):
        enum = node
        enum_name = enum.name if not enum.db_schema else f"{enum.db_schema}.{enum.name}"
        values = ",".join(f"'{value.value}'" for value in enum.values)
        return exp.Command(
            this="CREATE", expression=f"TYPE {enum_name} AS ENUM ({values})"
        )



from typing import List
from sqlglot import expressions as exp
from ...schema import Column, Enum, Reference, Table, TablePartial
from .base import BaseSQLConverter
from .column import ColumnConverter
from .index import IndexConverter, CompositePKIndexConverter
from .reference import ReferenceConverter


class TableConverter(BaseSQLConverter[Table]):
    def __init__(
        self,
        dialect,
        table_partials: List[TablePartial],
        references: List[Reference],
        enums: List[Enum],
    ):
        super().__init__(dialect)
        self.column_converter = ColumnConverter(dialect, enums)
        self.index_converter = IndexConverter(dialect, self)
        self.composite_pk_converter = CompositePKIndexConverter(dialect)
        self.reference_converter = ReferenceConverter(dialect)
        self.table_partials = table_partials
        self.references = references

    def convert(self, node):
        table = node
        # column name : column object
        column_map: dict[str, Column] = {}
        # Import all columns from table partials
        if table.table_partials:
            assert self.table_partials is not None
            for table_partial in table.table_partials:
                # First or default
                ref_table = next(
                    filter(
                        lambda table: table.name == table_partial, self.table_partials
                    ),
                    None,
                )
                assert ref_table is not None, (
                    f"Reference Table {table_partial} is not found"
                )
                if ref_table:
                    column_map.update(
                        map(lambda column: (column.name, column), ref_table.columns)
                    )
        # Columns in table will be added last
        # to override any overlapped columns
        # in the table partial list
        for column in table.columns:
            column_map.update(map(lambda column: (column.name, column), table.columns))

        # Column definitions
        col_defs = []
        for column_name in column_map:
            column = column_map[column_name]
            col_defs.append(self.column_converter.convert(column))

        # Constraints: Foreign Key
        constraint_defs = []
        for reference in self.references:
            constraint_defs.append(self.reference_converter.convert(reference))

        # Constraints: Primary Key
        if table.indexes:
            composite_pk_indexes = list(
                filter(
                    lambda index: index.settings and index.settings.is_primary_key,
                    table.indexes,
                )
            )
            for composite_pk_index in composite_pk_indexes:
                constraint_defs.append(
                    self.composite_pk_converter.convert(composite_pk_index)
                )

        table_def = exp.Create(
            this=exp.Schema(
                this=exp.Table(
                    this=exp.Identifier(this=table.name, quoted=True),
                    db=exp.Identifier(this=table.db_schema, quoted=True)
                    if table.db_schema
                    else None,
                ),
                expressions=col_defs + constraint_defs,
            ),
            kind="TABLE",
        )
        return table_def



from io import StringIO
import os
from sqlglot import Dialects, expressions as exp

from ...schema import Diagram
from .enum import EnumConverter
from .index import IndexConverter
from .table import TableConverter


def to_sql(diagram: Diagram, dialect: Dialects = Dialects.POSTGRES):
    endblock = ";" + os.linesep + os.linesep

    enum_converter = EnumConverter(dialect)
    with StringIO() as buffer:
        # Create schema
        schemas = set(
            map(
                lambda enum: enum.db_schema,
                filter(lambda enum: enum.db_schema is not None, diagram.enums),
            )
        )
        schemas |= set(
            map(
                lambda table: table.db_schema,
                filter(lambda table: table.db_schema is not None, diagram.tables),
            )
        )
        for schema in sorted(schemas):
            schema_def = exp.Create(
                this=exp.Table(db=exp.Identifier(this=schema, quoted=True)),
                kind="SCHEMA",
                exists=True,
            )
            buffer.write(schema_def.sql(dialect=dialect, pretty=True))
            buffer.write(endblock)

        # Only create enum if the dialect supports
        if dialect in [Dialects.POSTGRES, Dialects.DUCKDB]:
            # Create enum
            if diagram.enums:
                for enum in diagram.enums:
                    enum_def = enum_converter.convert(enum)
                    buffer.write(enum_def.sql(dialect, pretty=True))
                    buffer.write(endblock)

        # Create Table along with Indexes
        for table in diagram.tables:
            # Find any references of this table
            references = list(
                filter(
                    lambda ref: ref.from_table.db_schema == table.db_schema
                    and ref.from_table.name == table.name,
                    diagram.references,
                )
            )

            table_converter = TableConverter(
                dialect,
                table_partials=diagram.table_partials,
                references=references,
                enums=diagram.enums,
            )
            index_converter = IndexConverter(dialect, table)

            table_def = table_converter.convert(table)
            buffer.write(table_def.sql(dialect=dialect, pretty=True))
            buffer.write(endblock)

            if table.indexes:
                # Indexes that are not composite primary keys
                indexes = list(
                    filter(
                        lambda index: not (
                            index.settings and index.settings.is_primary_key
                        ),
                        table.indexes,
                    )
                )
                for index in indexes:
                    index_def = index_converter.convert(index)
                    buffer.write(index_def.sql(dialect=dialect, pretty=True))
                    buffer.write(endblock)

        return buffer.getvalue()



from functools import wraps
from lark import Token, Transformer, v_args
import logging
from .schema import (
    Diagram,
    Project,
    Enum,
    Note,
    Table,
    TableGroup,
    Reference,
    ReferenceSettings,
    TablePartial,
)

logger = logging.getLogger(__name__)


def log_transform(func):
    """
    A decorator to log the entry and exit of a Transformer method.
    """

    @wraps(func)
    def wrapper(self_transformer: Transformer, *args, **kwargs):
        rule_name = func.__name__
        logger.debug(f"Entering '{rule_name}' with args: {args}")

        # Call the original method
        result = func(self_transformer, *args, **kwargs)

        logger.debug(
            f"Exiting '{rule_name}'. Result: {result!r}"
        )  # !r for representation
        return result

    return wrapper


class DBMLTransformer(Transformer[Token, Diagram]):
    def __default__(self, data, children, _):
        return {data.value.strip(): children}

    # 1. Common values
    def IDENTIFIER(self, token):
        return token.value.strip("`")

    def STRING(self, token):
        return token.value[1:-1]

    def MULTILINE_STRING(self, token):
        return token.value[3:-3].strip()

    def NUMBER(self, token):
        if token.value.isnumeric():
            return int(token.value)
        else:
            try:
                return float(token.value)
            except ValueError:
                # If it can't be converted to a number, return as is
                return token.value

    def RELATIONSHIP(self, token):
        return token.value.strip()

    def REFERENTIAL_ACTION(self, token):
        return token.value.strip()

    def FUNC_EXP(self, token):
        return token.value

    def COLOR_HEX(self, token):
        return token.value.strip()

    def true(self, *_):
        return True

    def false(self, *_):
        return False

    def pair(self, kv):
        return kv

    def settings(self, pairs):
        return {"settings": dict(pairs)}

    def name(self, vars):
        if len(vars) == 1:
            return {"name": vars[0]}
        return {"db_schema": vars[0], "name": vars[1]}

    # ====== PROJECT ======
    @v_args(inline=True)
    @log_transform
    def project(self, name, *pairs) -> Project:
        data = name | dict(pairs)
        return Project.model_validate(data)

    # ====== STICKY NOTE & NOTE INLINE ======
    def note_inline(self, vars):
        return {"note": vars[0]}

    @log_transform
    def note(self, vars) -> Note:
        return Note.model_validate(vars[0] | {"note": vars[1]})

    # ====== ENUM ======
    def enum_value(self, vars):
        data = {"value": vars[0]}
        # Settings
        if len(vars) > 1:
            data.update(vars[1])
        return data

    @v_args(inline=True)
    @log_transform
    def enum(self, name, *enum_values) -> Enum:
        data = name | {"values": enum_values}
        return Enum.model_validate(data, by_alias=True)

    # ====== TABLE GROUP ======
    @v_args(inline=True)
    @log_transform
    def group(self, name, *vars) -> TableGroup:
        data = name | {"tables": []}
        for var in vars:
            if "name" in var:
                data["tables"].append(var)
            else:
                data.update(var)
        return TableGroup.model_validate(data)

    # ====== REFERENCE ======
    def ref_col(self, cols):
        return cols

    @v_args(inline=True)
    @log_transform
    def ref(self, name, ref_col, ref_inline):
        ref_inline["ref"]["from_table"] = name
        ref_inline["ref"]["from_columns"] = ref_col
        return ref_inline

    def ref_inline(self, vars):
        return {
            "ref": {"relationship": vars[0], "to_table": vars[1], "to_columns": vars[2]}
        }

    def ref_settings(self, vars):
        settings = {}
        for var in vars:
            if isinstance(var, tuple):
                settings[var[0]] = var[1]
            else:
                settings.update(var)
        return settings

    @log_transform
    def reference(self, vars) -> Reference:
        name_dict = {}
        settings = None
        # Has no name
        if len(vars) == 1:
            relationship = vars[0]
        elif len(vars) == 2:
            if "name" in vars[0]:
                name_dict = vars[0]
                relationship = vars[1]
            else:
                relationship = vars[0]
                settings = ReferenceSettings.model_validate(vars[1]["settings"])
        else:
            name_dict = vars[0]
            relationship = vars[1]
            settings = ReferenceSettings.model_validate(vars[2]["settings"])

        data = {
            "db_schema": name_dict.get("db_schema"),
            "name": name_dict.get("name"),
            "settings": settings,
        } | relationship["ref"]
        return Reference.model_validate(data)

    # ====== TABLE ======
    def is_primary_key(self, *_):
        return "is_primary_key", True

    def is_null(self, *_):
        return "is_null", True

    def is_not_null(self, *_):
        return "is_null", False

    def is_unique(self, *_):
        return "is_unique", True

    def is_increment(self, *_):
        return "is_increment", True

    def alias(self, vars):
        return {"alias": vars[0]}

    @v_args(inline=True)
    @log_transform
    def data_type(self, sql_type, *vars):
        return {
            "data_type": {
                "sql_type": sql_type,
                "length": int(vars[0].value) if len(vars) > 0 else None,
                "scale": int(vars[1].value) if len(vars) > 1 else None,
            }
        }

    def column_setting(self, pairs):
        return pairs[0]

    def column_settings(self, pairs_or_ref):
        settings = {}
        pairs = []
        for pair in pairs_or_ref:
            if isinstance(pair, dict):
                settings.update(pair)
            else:
                pairs.append(pair)
        if pairs:
            settings.update(dict(pairs))
        return {"settings": settings}

    @v_args(inline=True)
    @log_transform
    def column(self, name, column_type, *settings):
        if "data_type" not in column_type:
            column_type = {"data_type": column_type}
        data = {"name": name} | column_type
        if settings:
            data.update(settings[0])
        return {"column": data}

    @v_args(inline=True)
    @log_transform
    def index(self, columns, *settings):
        # index_exp rule
        if isinstance(columns, dict):
            data = {"columns": columns["index_exp"]}
        else:
            data = {"columns": columns}
        if settings:
            data.update(settings[0])
        return data

    @v_args(inline=True)
    @log_transform
    def table_partial(self, name, *vars) -> TablePartial:
        data = name | {"columns": []}
        for var in vars:
            if "column" in var:
                data["columns"].append(var["column"])
            else:
                data.update(var)
        return TablePartial.model_validate(data)

    @v_args(inline=True)
    @log_transform
    def table(self, name, *vars) -> Table:
        data = name | {"columns": []}
        for var in vars:
            if "column" in var:
                data["columns"].append(var["column"])
            elif isinstance(var, str):
                table_partials = data.get("table_partials", [])
                table_partials.append(var)
                data["table_partials"] = table_partials
            else:
                data.update(var)
        return Table.model_validate(data)

    # ====== DIAGRAM ======
    def start(self, items) -> Diagram:
        diagram = Diagram.model_construct()
        for item in items:
            if isinstance(item, Project):
                diagram.project = item
            elif isinstance(item, Enum):
                diagram.enums.append(item)
            elif isinstance(item, Reference):
                diagram.references.append(item)
            elif isinstance(item, TableGroup):
                diagram.table_groups.append(item)
            elif isinstance(item, Note):
                diagram.sticky_notes.append(item)
            elif isinstance(item, Table):
                diagram.tables.append(item)
            elif isinstance(item, TablePartial):
                diagram.table_partials.append(item)

        return diagram
